from typing import Optional

import jwt
import requests


TOKEN = requests.get("https://public-keys.teamtv.nl/app.teamtv.nl.pub").content

DEFAULT_LEEWAY = 15 * 60


def decode(jwt_token: str, app_id: Optional[str] = None, verify: bool = True):
    options = None
    if not verify:
        options = {"verify_signature": False}

    audience = None
    if app_id:
        audience = f"app:{app_id}"

    return jwt.decode(
        jwt_token,
        TOKEN,
        algorithms="RS256",
        verify=verify,
        audience=audience,
        options=options,
        leeway=DEFAULT_LEEWAY,
    )



class PyTeamTVError(Exception):
    pass


class TeamNotFound(PyTeamTVError):
    pass


class InputError(PyTeamTVError):
    pass


class TokenMissing(PyTeamTVError):
    pass


class ConfigurationError(PyTeamTVError):
    pass



from pyteamtv.models.sharing_group import SharingGroup
from . import _ResourceGroup

from .capabilities import _HasTeamsMixin, _HasSportingEventsMixin, _HasVideosMixin


class SharingGroupResourceGroup(
    _ResourceGroup, _HasTeamsMixin, _HasSportingEventsMixin, _HasVideosMixin
):
    def get_sharing_group(self):
        return SharingGroup(
            self._requester,
            self._requester.request("GET", f"/sharingGroups/{self.sharing_group_id}"),
        )

    @property
    def sharing_group_id(self):
        return self._sharing_group_id

    def _use_attributes(self, attributes: dict):
        self._sharing_group_id = attributes["targetResourceId"].split(":")[1]
        super()._use_attributes(attributes)



from datetime import timedelta, datetime
from typing import List

from .event_store import Event

from ...models.match.match_state import MatchState


def diff(end_time, start_time) -> float:
    d = end_time - start_time
    if isinstance(d, timedelta):
        return d.total_seconds()
    return d


def calculate_match_state(events: List[Event], timestamp: datetime) -> MatchState:
    if len(events) == 0:
        return MatchState(
            home_score=0, away_score=0, current_period=0, period_active=False, time=0
        )

    home_team_id = None
    away_team_id = None

    home_score = 0
    away_score = 0

    time_correction = 0
    pause_time = None

    period_active = False
    current_period = 0
    period_start_time = 0

    current_possession_team_id = None

    for event in events:
        event_type = event.event_type
        event_attributes = event.attributes
        occurred_on = event.occurred_on

        if event_type == "SportingEventCreated":
            home_team_id = event_attributes["homeTeam"]["teamId"]
            away_team_id = event_attributes["awayTeam"]["teamId"]
        elif event_type == "Goal":
            if event_attributes["goalAttributes"]["teamId"] == home_team_id:
                home_score += 1
            elif event_attributes["goalAttributes"]["teamId"] == away_team_id:
                away_score += 1

        elif event_type == "StartPossession":
            current_possession_team_id = event_attributes["startPossessionAttributes"][
                "teamId"
            ]
        elif event_type == "Shot":
            shot_attributes = event_attributes["shotAttributes"]
            if shot_attributes["result"] == "GOAL":
                if current_possession_team_id == home_team_id:
                    home_score += 1
                elif current_possession_team_id == away_team_id:
                    away_score += 1

        elif event_type == "StartPeriod":
            period_start_time = occurred_on
            time_correction = 0
            pause_time = None
            period_active = True
            current_period = int(event_attributes["period"])
        elif event_type == "Time:pause":
            # dont break when events are entered multiple times
            if pause_time is None:
                pause_time = occurred_on
        elif event_type == "Time:resume":
            # dont break when events are entered multiple times
            if pause_time is not None:
                time_correction -= diff(occurred_on, pause_time)
                pause_time = None
        elif event_type == "Time:correction":
            time_correction += float(
                event_attributes["time:correctionAttributes"]["seconds"]
            )
        elif event_type == "EndPeriod":
            period_active = False

    if period_active:
        if pause_time is None:
            base_time = diff(timestamp, period_start_time)
        else:
            base_time = diff(pause_time, period_start_time)
        base_time += time_correction

        current_time = base_time
    else:
        current_time = 0

    return MatchState(
        home_score=home_score,
        away_score=away_score,
        time=current_time,
        current_period=current_period,
        period_active=period_active,
        current_possession_team_id=current_possession_team_id,
    )



from dataclasses import dataclass

from .teamtv_object import TeamTVObject


@dataclass
class Credentials(object):
    access_key_id: str
    secret_access_key: str
    session_token: str


@dataclass
class BucketInfo(object):
    bucket: str
    base_key: str


class AppStorageTokens(TeamTVObject):
    @property
    def credentials(self):
        return self._credentials

    def _use_attributes(self, attributes: dict):
        self._credentials = Credentials(
            access_key_id=attributes["credentials"]["accessKeyId"],
            secret_access_key=attributes["credentials"]["secretAccessKey"],
            session_token=attributes["credentials"]["sessionToken"],
        )
        self._bucket_info = BucketInfo(
            bucket=attributes["storage"]["bucket"],
            base_key=attributes["storage"]["baseKey"],
        )

        super()._use_attributes(attributes)



from abc import ABC, abstractmethod
from queue import Queue
from typing import List, Dict
import logging


class Event:
    def __init__(self, event_type: str, occurred_on, attributes: Dict):
        self.event_type = event_type
        self.occurred_on = occurred_on
        self.attributes = attributes

        event_attribute_key = event_type[0].lower() + event_type[1:] + "Attributes"
        if not self.attributes.get(event_attribute_key):
            self.attributes[event_attribute_key] = {}

    def __repr__(self):
        return f'<Event type="{self.event_type}" occurred_on="{self.occurred_on}" attributes={self.attributes}>'


class EventStore(ABC):
    @abstractmethod
    def get_events(self, timestamp) -> List[Event]:
        pass


class QueueEventStore(EventStore):
    def __init__(self, match_event_queue: Queue):
        self._match_event_queue = match_event_queue
        self._events = []  # type: List[Event]

    def _update_events(self):
        while not self._match_event_queue.empty():
            event = self._match_event_queue.get()
            logging.debug("Received event: {}".format(event))
            event_type = event["eventType"]
            event_attributes = event["eventAttributes"]
            occurred_on = event["occurredOn"]

            if event_type == "SportingEventCreated":
                self._events = []

            if event_type in ("ObservationRemoved", "SynchronizationPointRemoved"):
                # It is possible for a SynchronizationPointRemoved event to match
                # multiple StartPeriod events. This is because the id is "START_PERIOD:1" etc.
                # When a SynchronizationPoint is removed, it will probably be re-added later with the same
                # id. Make sure we only remove events BEFORE the SynchronizationPointRemoved event, otherwise
                # we will remove events in the future that should not be removed.
                event_to_remove_id = event_attributes["id"]
                indexes_to_remove = []
                for i, event_ in enumerate(self._events):
                    event_attributes_ = event_.attributes
                    if event_attributes_.get("id") == event_to_remove_id:
                        indexes_to_remove.append(i)
                        break

                for i in reversed(indexes_to_remove):
                    # print("removing", event_to_remove_id)
                    # print(self._events[i])
                    del self._events[i]
            else:
                self._events.append(
                    Event(
                        event_type=event_type,
                        attributes=event_attributes,
                        occurred_on=occurred_on,  # datetime.fromtimestamp(float(event_attributes['time']), timezone.utc) if 'time' in event_attributes else occurred_on
                    )
                )

    def get_events(self, timestamp=None) -> List[Event]:
        self._update_events()

        def _filter(event: Event):
            if event.event_type == "SportingEventCreated":
                return True
            return event.occurred_on <= timestamp

        if timestamp:
            return list(filter(_filter, self._events))
        else:
            return self._events


class StaticEventStore(EventStore):
    def __init__(self, events: List[Dict]):
        self._events = [
            Event(
                event_type=event["eventType"],
                attributes=event["eventAttributes"],
                occurred_on=event["occurredOn"],
            )
            for event in events
        ]

    def get_events(self, timestamp) -> List[Event]:
        def _filter(event: Event):
            if event.event_type == "SportingEventCreated":
                return True
            return event.occurred_on <= timestamp

        return list(filter(_filter, self._events))



from dataclasses import dataclass
from typing import Optional


@dataclass
class Team:
    team_id: str
    name: str
    short_code: str
    primary_color: Optional[str] = None
    secondary_color: Optional[str] = None


@dataclass
class MatchConfig:
    home_team: Team
    away_team: Team
    period_count: int

    @classmethod
    def empty(cls):
        return cls(
            home_team=Team(team_id="home", name="Home", short_code="HOM"),
            away_team=Team(team_id="away", name="Away", short_code="AWA"),
            period_count=2,
        )



from pyteamtv.infra.requester import Requester
from pyteamtv.models.resource_group.factory import factory as resource_group_factory
from .endpoint import API_ENDPOINT

from .token import decode
from ..models.access_requester import AccessRequester


class TeamTVApp(object):
    def __init__(self, jwt_token, app_id: str, use_cache: bool = False):
        self.jwt_token = jwt_token

        token = decode(jwt_token, app_id)

        self._requester = Requester(
            f"{API_ENDPOINT}/api", jwt_token, use_cache=use_cache
        )
        self.__token = token

    def get_resource_group(self):
        """
        :rtype: :class:`pyteamtv.resource_group.team.Team`
        """
        data = self._requester.request("GET", "/resourceGroups/current")

        return resource_group_factory(self._requester, data)

    def get_access_requester(self) -> AccessRequester:
        return AccessRequester(
            user_id=self.__token["sub"],
            tenant_id=self.__token["scope"]["tenantId"],
            role_names=self.__token["scope"]["roleNames"],
            resource_group_id=self.__token["scope"]["resourceGroupId"],
        )
